(*#######################################################################################

UpdateLoader - Hauptprogramm

Standardeinstellungen können in protocol.pas angepasst werden

http://www.leo-andres.de/

Copyright (C) 2012 Leo-Andres Hofmann


This file is part of UpdateLoader.

UpdateLoader is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

UpdateLoader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with UpdateLoader. If not, see <http://www.gnu.org/licenses/>.

#######################################################################################*)

unit main;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ComCtrls, FileInfo,
  Grids, StdCtrls, ExtCtrls, EditBtn, Buttons, LCLType, IniFiles, LCLIntf, Tools, Protocol, SynaSer;

type

  { TMainWindow }

  TMainWindow = class(TForm)
    btnClearSettings: TButton;
    btnUpdateControl: TButton;
    btnSelectFirmwareFilename: TButton;
    ckbProgramStart: TCheckBox;
    ckbShowMemory: TCheckBox;
    ckbReloadFile: TCheckBox;
    ckbProgramWrite: TCheckBox;
    ckbProgramVerify: TCheckBox;
    ckbHideMessages: TCheckBox;
    ckbAdvancedMode: TCheckBox;
    cmbComPort: TComboBox;
    cmbBaudRate: TComboBox;
    gpbMemoryView: TGroupBox;
    gpbStatus: TGroupBox;
    gpbAdvancedSettings: TGroupBox;
    gpbConnection: TGroupBox;
    gpbUpdate: TGroupBox;
    gpbInfo: TGroupBox;
    lblProgramConnect: TLabel;
    lblProgramWrite: TLabel;
    lblProgramVerify: TLabel;
    lblUpdateProgress: TLabel;
    lblFlashNotice: TLabel;
    lblBaudRate: TLabel;
    lblComPort: TLabel;
    edtBootloaderPassword: TLabeledEdit;
    edtFirmwareFilename: TLabeledEdit;
    edtConnectTries: TLabeledEdit;
    memUpdateStatus: TMemo;
    memUpdateLog: TMemo;
    OpenDialog: TOpenDialog;
    PageControl: TPageControl;
    pgbProgramConnect: TProgressBar;
    pgbProgramWrite: TProgressBar;
    pgbUpdateProgress: TProgressBar;
    pgbProgramVerify: TProgressBar;
    StatusBar: TStatusBar;
    stgMemoryView: TStringGrid;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    tmrFlashNotice: TTimer;
    procedure btnSelectFirmwareFilenameClick(Sender: TObject);
    procedure btnClearSettingsClick(Sender: TObject);
    procedure btnUpdateControlClick(Sender: TObject);
    procedure ckbAdvancedModeChange(Sender: TObject);
    procedure ckbShowMemoryChange(Sender: TObject);
    procedure edtConnectTriesChange(Sender: TObject);
    procedure edtFirmwareFilenameChange(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCloseQuery(Sender: TObject; var CanClose: boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure PageControlChanging(Sender: TObject; var AllowChange: Boolean);
    procedure stgMemoryViewDrawCell(Sender: TObject; aCol, aRow: Integer;
      aRect: TRect; aState: TGridDrawState);
    procedure TabSheet2Show(Sender: TObject);
    procedure tmrFlashNoticeTimer(Sender: TObject);
  private
    { private declarations }
  var
    updateBuffer: TMemoryBuffer;
    updateBufferValid: boolean;
    updateBufferUsed: longint;

    memoryViewUpdate: boolean;
    programmingActive: boolean;

    settingsFile: TMemIniFile;

    procedure logMessage(message: string; openMessageBox: boolean = false; addToLog: boolean = true; windowIcon: integer = MB_ICONEXCLAMATION);
    procedure infoMessage(message: string);
    procedure flashNotice(showNotice: boolean; message: string = ''; flashSpeed: integer = 500);
    procedure sleepActive(duration: integer);
    function doFirmwareUpdate(serialConnection: TBlockSerial; bootloaderRoutines: TBootloaderRoutines; programWrite, programVerify, programStart, showMessages: boolean): boolean;
  public
    { public declarations }
  end; 

var
  MainWindow: TMainWindow;

implementation

{$R *.lfm}

{ TMainWindow }
{$REGION 'Programmstart/Ende'}
procedure TMainWindow.FormShow(Sender: TObject);
var
  comportsCount: integer;
  selectedPage: integer;
  fileVersion:  TFileVersionInfo;
  itemIndex: integer;
begin
  try
    comportsCount := getComportCount();
  except
  end;

  if comportsCount <= 0 then
  begin
    logMessage('Dieser Computer verfügt über keine serielle Schnittstelle (COM-Port). "' +
      sLineBreak + 'Mindestens eine Schnittstelle ist notwendig um dieses Programm zu verwenden.' +
      sLineBreak + sLineBreak + 'Das Programm wird jetzt beendet.', true, false, MB_ICONSTOP);

    MainWindow.Hide;
    Application.Terminate;
  end;

  //Fenstertitel
  try
    try
      fileVersion := TFileVersionInfo.Create(Application);
      fileVersion.fileName := Application.ExeName;

      MainWindow.Caption := fileversion.getVersionSetting('ProductName') + ' ' +
        fileversion.getVersionSetting('ProductVersion') + ' (Build ' +
        fileversion.getVersionSetting('FileVersion') + ')';
    finally
      fileVersion.Free;
    end;
  except
  end;

  try
    getComportNames(cmbComPort.Items);
  except
  end;

  cmbComPort.ItemIndex := 0;
  cmbBaudRate.ItemIndex := defaultBaudrate;
  StatusBar.Panels[0].Text := 'Bereit, keine Firmware-Datei geladen';

  try
    settingsFile := TMemIniFile.Create(ExtractFilePath(Application.ExeName) + 'UpdateLoader.ini');

    if Assigned(settingsFile) then
    begin
      itemIndex := settingsFile.ReadInteger('UpdateLoader', 'ComPort', 0);
      if (itemIndex <= comportsCount) and (itemIndex >= 0) then
         cmbComPort.ItemIndex := itemIndex;

      itemIndex := settingsFile.ReadInteger('UpdateLoader', 'BaudRate', defaultBaudrate);
      if (itemIndex <= cmbBaudRate.Items.Count) and (itemIndex >= 0) then
         cmbBaudRate.ItemIndex := itemIndex;

      ckbAdvancedMode.Checked := settingsFile.ReadBool('UpdateLoader', 'AdvancedMode', false);
      ckbProgramWrite.Checked := settingsFile.ReadBool('UpdateLoader', 'ProgramWrite', true);
      ckbProgramVerify.Checked := settingsFile.ReadBool('UpdateLoader', 'ProgramVerify', true);
      ckbProgramStart.Checked := settingsFile.ReadBool('UpdateLoader', 'ProgramStart', true);
      ckbReloadFile.Checked := settingsFile.ReadBool('UpdateLoader', 'ReloadFile', true);
      ckbHideMessages.Checked := settingsFile.ReadBool('UpdateLoader', 'HideMessages', false);
      ckbShowMemory.Checked := settingsFile.ReadBool('UpdateLoader', 'ShowMemory', false);

      edtBootloaderPassword.Text := settingsFile.ReadString('UpdateLoader', 'BootloaderPassword', defaultBootloaderPassword);
      edtFirmwareFilename.Text := settingsFile.ReadString('UpdateLoader', 'FirmwareFilename', '');

      if settingsFile.ReadInteger('UpdateLoader', 'ConnectTries', defaultConnectTries) > 0 then
         edtConnectTries.Text := IntToStr(settingsFile.ReadInteger('UpdateLoader', 'ConnectTries', defaultConnectTries))
      else
         edtConnectTries.Text := IntToStr(defaultConnectTries);

      selectedPage := settingsFile.ReadInteger('UpdateLoader', 'Page', 0);
    end;
  except
    on Exception do
       ShowMessage('Fehler beim Laden der Programmeinstellungen');
  end;

  ckbAdvancedModeChange(Self);

  if (TabSheet2.TabVisible = true) and (selectedPage = 1) then
     PageControl.TabIndex := 1
  else
    PageControl.TabIndex := 0;

end;

procedure TMainWindow.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  try
     try
        if Assigned(settingsFile) then
        begin
          settingsFile.Clear;

          settingsFile.WriteInteger('UpdateLoader', 'ComPort', cmbComPort.ItemIndex);
          settingsFile.WriteInteger('UpdateLoader', 'BaudRate', cmbBaudRate.ItemIndex);

          settingsFile.WriteBool('UpdateLoader', 'AdvancedMode', ckbAdvancedMode.Checked);
          settingsFile.WriteBool('UpdateLoader', 'ProgramWrite', ckbProgramWrite.Checked);
          settingsFile.WriteBool('UpdateLoader', 'ProgramVerify', ckbProgramVerify.Checked);
          settingsFile.WriteBool('UpdateLoader', 'ProgramStart', ckbProgramStart.Checked);
          settingsFile.WriteBool('UpdateLoader', 'ReloadFile', ckbReloadFile.Checked);
          settingsFile.WriteBool('UpdateLoader', 'HideMessages', ckbHideMessages.Checked);
          settingsFile.WriteBool('UpdateLoader', 'ShowMemory', ckbShowMemory.Checked);

          settingsFile.WriteString('UpdateLoader', 'BootloaderPassword', edtBootloaderPassword.Text);
          settingsFile.WriteString('UpdateLoader', 'FirmwareFilename', edtFirmwareFilename.Text);

          if StrToInt(edtConnectTries.Text) > 0 then
             settingsFile.WriteInteger('UpdateLoader', 'ConnectTries', StrToInt(edtConnectTries.Text))
          else
             settingsFile.WriteInteger('UpdateLoader', 'ConnectTries', defaultConnectTries);

          settingsFile.WriteInteger('UpdateLoader', 'Page', PageControl.TabIndex);

          settingsFile.UpdateFile;
        end;
        finally
          settingsFile.Free;
        end;
  except
    on Exception do
       ShowMessage('Fehler beim Speichern der Programmeinstellungen');
  end;

  CloseAction := caFree;
end;

procedure TMainWindow.FormCloseQuery(Sender: TObject; var CanClose: boolean);
begin
  CanClose := not programmingActive;
end;

procedure TMainWindow.FormCreate(Sender: TObject);
begin
  inherited;

  SetLength(updateBuffer, settingMaximumBufferSize);
  clearBuffer(updateBuffer);

  updateBufferValid := false;
  updateBufferUsed := 0;

  memoryViewUpdate := false;
  programmingActive := false;
end;
{$ENDREGION}

{$REGION 'Events (Button-Betätigung)'}
procedure TMainWindow.PageControlChanging(Sender: TObject;
  var AllowChange: Boolean);
begin
  if (programmingActive = true) and (PageControl.PageIndex = 0) then
    PageControl.PageIndex := 0;

  AllowChange := not programmingActive;
end;

procedure TMainWindow.btnClearSettingsClick(Sender: TObject);
var
  reply: integer;
begin
  reply := Application.MessageBox('Einstellungen wirklich zurücksetzen?', 'UpdateLoader', MB_ICONQUESTION or MB_YESNO or MB_DEFBUTTON2);

  if reply = IDYES then
  begin
    cmbComPort.ItemIndex := 0;
    cmbBaudRate.ItemIndex := defaultBaudrate;

    ckbAdvancedMode.Checked := false;
    ckbProgramWrite.Checked := true;
    ckbProgramVerify.Checked := true;
    ckbProgramStart.Checked := true;
    ckbReloadFile.Checked := true;
    ckbHideMessages.Checked := false;
    ckbShowMemory.Checked := false;

    edtBootloaderPassword.Text := defaultBootloaderPassword;
    edtFirmwareFilename.Text := '';

    edtConnectTries.Text := IntToStr(defaultConnectTries);

    ckbAdvancedModeChange(Self);
    edtFirmwareFilenameChange(Self);
  end;
end;

procedure TMainWindow.btnSelectFirmwareFilenameClick(Sender: TObject);
begin
  if edtFirmwareFilename.Text <> '' then
     OpenDialog.Filename := edtFirmwareFilename.Text
  else
     OpenDialog.Filename := ExtractFilePath(Application.ExeName);

  if OpenDialog.Execute then
  begin
    if edtFirmwareFilename.Text = OpenDialog.FileName then
      edtFirmwareFilenameChange(Self)
    else
      edtFirmwareFilename.Text := OpenDialog.FileName;
  end;
end;

procedure TMainWindow.ckbAdvancedModeChange(Sender: TObject);
begin
  TabSheet2.TabVisible := ckbAdvancedMode.Checked;
  if (TabSheet2.TabVisible = false) and (PageControl.TabIndex = 1) then
     PageControl.TabIndex := 0;
end;

procedure TMainWindow.ckbShowMemoryChange(Sender: TObject);
begin
  if (ckbShowMemory.Checked = true) and (updateBufferValid = true) then
  begin
    memoryViewUpdate := true;

    if PageControl.TabIndex = 1 then
       TabSheet2Show(Self);
  end
  else
  begin
    memoryViewUpdate := false;

    stgMemoryView.RowCount := 0;
    gpbMemoryView.Caption := ' Programm ';
  end;
end;

procedure TMainWindow.edtConnectTriesChange(Sender: TObject);
begin
  try
    StrToInt(edtConnectTries.Text);
  except
    edtConnectTries.Text := IntToStr(defaultConnectTries);
    exit;
  end;
  if StrToInt(edtConnectTries.Text) <= 0 then
     edtConnectTries.Text := IntToStr(defaultConnectTries);
end;

procedure TMainWindow.edtFirmwareFilenameChange(Sender: TObject);
var
  success: boolean;
  oldCursor: TCursor;
begin
  success := false;

  if (FileExists(edtFirmwareFilename.Text)) and (length(edtFirmwareFilename.Text) > 0) then
  begin
    oldCursor := Screen.Cursor;
    Screen.Cursor := crHourGlass;
    pgbUpdateProgress.Style := pbstMarquee;

    updateBufferUsed := 0;
    try
      success := loadProgramFile(edtFirmwareFilename.Text, updateBuffer, settingMaximumBufferSize, updateBufferUsed);
    except
      on E : Exception do
         logMessage(E.Message + ' (' + ExtractFileName(edtFirmwareFilename.Text) + ')', true, true);
    end;

    if success = true then
    begin
      updateBufferValid := true;

      ckbShowMemoryChange(Self);

      logMessage('Firmware-Datei geladen (' + ExtractFileName(edtFirmwareFilename.Text) + ')');
      StatusBar.Panels[0].Text := 'Bereit, Firmware-Datei geladen (' + ExtractFileName(edtFirmwareFilename.Text) + ')';
      btnUpdateControl.Enabled := true;
    end
    else
    begin
      updateBufferValid := false;
      btnUpdateControl.Enabled := false;
      StatusBar.Panels[0].Text := 'Bereit, keine Firmware-Datei geladen';

      ckbShowMemoryChange(Self);
    end;

    Screen.Cursor := oldCursor;
    pgbUpdateProgress.Style := pbstNormal;
  end
  else
  begin
    updateBufferValid := false;
    btnUpdateControl.Enabled := false;
    StatusBar.Panels[0].Text := 'Bereit, keine Firmware-Datei geladen';

    edtFirmwareFilename.Text := '';

    ckbShowMemoryChange(Self);
  end;
end;

procedure TMainWindow.stgMemoryViewDrawCell(Sender: TObject; aCol,
  aRow: Integer; aRect: TRect; aState: TGridDrawState);
begin
  if gdFixed in aState then
     Exit;

  case aCol of
  0:
    begin
      stgMemoryView.Canvas.Font.Color := RGB(128, 0, 0);
    end;
  4:
    begin
      stgMemoryView.Canvas.Font.Color := RGB(128, 0, 128);
    end;
  else
    begin
      stgMemoryView.Canvas.Font.Color := RGB(0, 0, 128);
    end;
  end;

  stgMemoryView.Canvas.TextOut(aRect.Left + 2, aRect.Top + 2, stgMemoryView.Cells[aCol, aRow]);
  stgMemoryView.Canvas.TextRect(aRect, aRect.Left + 2, aRect.Top + 2, stgMemoryView.Cells[aCol, aRow]);
end;

procedure TMainWindow.TabSheet2Show(Sender: TObject);
var
  oldCursor: TCursor;
begin
  if memoryViewUpdate = true then
  begin
    memoryViewUpdate := false;

    oldCursor := Screen.Cursor;
    Screen.Cursor := crHourGlass;

    showProgramMemory(stgMemoryView, updateBuffer, updateBufferUsed, pgbUpdateProgress);
    gpbMemoryView.Caption := ' Programm (' + convertKB2String(updateBufferUsed) + ') ';

    Screen.Cursor := oldCursor;
  end;
end;
{$ENDREGION}

{$REGION 'Hilfsfunktionen für die Benutzerführung'}
procedure TMainWindow.logMessage(message: string; openMessageBox: boolean = false; addToLog: boolean = true; windowIcon: integer = MB_ICONEXCLAMATION);
begin
  if addToLog = true then
    memUpdateLog.Lines.Add(message);

  if openMessageBox = true then
     Application.MessageBox(PChar(message), 'UpdateLoader - Nachricht', MB_OK + windowIcon);
end;

procedure TMainWindow.infoMessage(message: string);
begin
  memUpdateStatus.Lines.Add(message);
end;

procedure TMainWindow.flashNotice(showNotice: boolean; message: string = ''; flashSpeed: integer = 500);
begin
  lblFlashNotice.Caption := message;
  lblFlashNotice.Visible := showNotice;
  lblFlashNotice.Left := round((gpbInfo.Width - lblFlashNotice.Width) / 2);

  tmrFlashNotice.Interval := flashSpeed;
  tmrFlashNotice.Enabled := showNotice;
end;

procedure TMainWindow.tmrFlashNoticeTimer(Sender: TObject);
begin
  lblFlashNotice.Visible := not lblFlashNotice.Visible;
end;

procedure TMainWindow.sleepActive(duration: integer);
var
  i: integer;
begin
  if duration <= 0 then
     Exit;

  for i := 0 to Round(duration / 100) do
  begin
       Sleep(100);
       Application.ProcessMessages;
  end;
end;
{$ENDREGION}

{$REGION 'Update-Routine'}
procedure TMainWindow.btnUpdateControlClick(Sender: TObject);
var
  programWrite, programVerify, programStart, showMessages: boolean;
  reloadFile: boolean;
  reply: integer;
  startTime: TDateTime;
  oldCursor: TCursor;
  updateDuration: string;
  oldStatus: string;
  success: boolean;
  serialConnection: TBlockSerial;
  bootloaderRoutines: TBootloaderRoutines;
begin
  showMessages := not (ckbHideMessages.Checked and ckbAdvancedMode.Checked);
  programWrite := ckbProgramWrite.Checked or (not ckbAdvancedMode.Checked);
  programVerify := ckbProgramVerify.Checked or (not ckbAdvancedMode.Checked);
  programStart := ckbProgramStart.Checked or (not ckbAdvancedMode.Checked);
  reloadFile := ckbReloadFile.Checked or (not ckbAdvancedMode.Checked);

  if programmingActive = true then
    Exit;

  if (programWrite = false) and (programVerify = false) then
  begin
    logMessage('Mindestens einen Schritt (Programmieren/Überprüfen) auswählen.' +
      sLineBreak + sLineBreak + 'Nichts zu tun!', true, false, MB_ICONINFORMATION);
    Exit;
  end;

  if reloadFile = true then
    edtFirmwareFilenameChange(Self);

  if updateBufferValid = false then
    Exit;

  if showMessages = true then
  begin
    reply := Application.MessageBox(
          'Warnung: Die Aktualisierung der Firmware kann nicht mehr rückgängig gemacht werden.' +
          ' Im Fehlerfall kann das Gerät beschädigt werden.' +
          sLineBreak + sLineBreak + sLineBreak + 'Update wirklich durchführen?', 'UpdateLoader', MB_ICONWARNING or MB_YESNO or MB_DEFBUTTON2);
    if reply = IDNO then
       Exit;
  end;

  //Initialisierung
  programmingActive := true;
  startTime := TDateTime(now);

  serialConnection := TBlockSerial.Create;
  bootloaderRoutines := TBootloaderRoutines.Create;

  //GUI sperren
  btnSelectFirmwareFilename.Enabled := false;
  btnUpdateControl.Enabled := false;
  cmbComPort.Enabled := false;
  edtBootloaderPassword.Enabled := false;
  edtFirmwareFilename.Enabled := false;
  ckbAdvancedMode.Enabled := false;
  PageControl.PageIndex := 0;

  //Statusanzeigen setzen
  oldCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;
  oldStatus := StatusBar.Panels[0].Text;
  memUpdateStatus.Clear;

  if showMessages = true then
    MainWindow.FormStyle := fsStayOnTop;

  btnUpdateControl.Caption := 'Bitte warten...';
  StatusBar.Panels[0].Text := 'Firmware-Update wird ausgeführt, Benutzeroberfläche ist deaktiviert!';
  logMessage('Firmware-Update gestartet (' + convertKB2String(updateBufferUsed) + ')');

  //Update ausführen
  success := doFirmwareUpdate(serialConnection, bootloaderRoutines, programWrite, programVerify, programStart, showMessages);

  //Verbindung freigeben
  if serialConnection.InstanceActive = true then
  begin
    serialConnection.Purge;
    serialConnection.Flush;
    serialConnection.CloseSocket;
  end;

  //Objekte löschen
  serialConnection.Free;
  bootloaderRoutines.Free;

  //GUI freigeben
  btnSelectFirmwareFilename.Enabled := true;
  btnUpdateControl.Enabled := true;
  cmbComPort.Enabled := true;
  edtBootloaderPassword.Enabled := true;
  edtFirmwareFilename.Enabled := true;
  ckbAdvancedMode.Enabled := true;

  //Statusanzeige zurücksetzen
  flashNotice(false);

  Screen.Cursor := oldCursor;
  StatusBar.Panels[0].Text := oldStatus;

  if showMessages = true then
    MainWindow.FormStyle := fsNormal;

  btnUpdateControl.Caption := 'Update starten!';
  pgbProgramConnect.Position := 0;
  pgbProgramWrite.Position := 0;
  pgbProgramVerify.Position := 0;
  pgbUpdateProgress.Position := 0;

  //Meldung
  if success = true then
  begin
    logMessage('Firmware-Update erfolgreich beendet!');

    DateTimeToString(updateDuration, 'hh:mm:ss', TDateTime(now) - startTime);
    logMessage('Gesamtdauer: ' + updateDuration);
  end
  else
  begin
    infoMessage('Fehler aufgetreten, Update abgebrochen!');
    logMessage('Fehler beim Firmware-Update, Ausführung abgebrochen');
  end;

  //Status zurücksetzen, Freigabe für nächsten Durchlauf
  programmingActive := false;
end;

function TMainWindow.doFirmwareUpdate(serialConnection: TBlockSerial; bootloaderRoutines: TBootloaderRoutines; programWrite, programVerify, programStart, showMessages: boolean): boolean;
var
  comPort: string;
  wireMode: integer;
  oneWire: boolean;
  connectTries: integer;
  reply: integer;
  supportsCRC: boolean;
  supportsVerify: boolean;
  deviceName: string;
  deviceFlashSize: longint;
  deviceWriteBuffer: integer;
  deviceRevision: string;
  baudRate: integer;
begin
  oneWire := false;
  supportsCRC := false;
  supportsVerify := false;
  Result := false;

  pgbUpdateProgress.Position := 0;
  pgbUpdateProgress.Max := 60;
  pgbUpdateProgress.Step := 10;

  infoMessage('COM-Port öffnen');

  //Port und Einstellungen ermitteln
  comPort := LeftStr(cmbComPort.Text, Pos(':', cmbComPort.Text) - 1);

  baudRate := StrToInt(cmbBaudRate.Text);
  if (baudRate <= 0) or (ckbAdvancedMode.Checked = false) then
    baudRate := StrToInt(cmbBaudRate.Items[defaultBaudrate]);

  //Schnittstelle konfigurieren
  serialConnection.RaiseExcept := true;
  serialConnection.LinuxLock := false;
  serialConnection.InterPacketTimeout := false;

  //Com-Port öffnen
  try
    serialConnection.Connect(comPort);
    sleepActive(settingPortSetupDelay);
    serialConnection.Config(baudRate, 8, 'N', SB1, false, false);
    sleepActive(settingPortSetupDelay);
    serialConnection.Purge;
    serialConnection.Flush;
  except
    on E : Exception do
    begin
      logMessage(serialConnection.GetErrorDesc(serialConnection.LastError) +
        sLineBreak + sLineBreak + 'COM-Port konnte nicht geöffnet werden.' +
        sLineBreak + 'Port wird eventuell von anderer Anwendung blockiert', true, false, MB_ICONSTOP);
      Exit;
    end;
  end;

  //Geöffnete Verbindung an Bootloader-Klasse übergeben
  bootloaderRoutines.assignSerialConnection(serialConnection);

  pgbUpdateProgress.StepIt;
  Application.ProcessMessages;
  infoMessage('Verbindungsaufbau...');

  //Benutzerführung
  if showMessages = true then
  begin
    logMessage('Gerät jetzt ausschalten und weitere Anweisungen abwarten.' +
      sLineBreak + sLineBreak + 'Verbindung nicht trennen!', true, false, MB_ICONINFORMATION);

    sleepActive(settingConnectMessageDelay);
    flashNotice(true, 'Gerät wieder einschalten!', 500);
    sleepActive(settingConnectMessageDelay);
  end
  else
  begin
    flashNotice(true, 'Bootloader aktivieren!', 500);
    sleepActive(settingConnectMessageDelay);
  end;

  //Verbindung herstellen
  pgbUpdateProgress.StepIt;
  Application.ProcessMessages;

  try
    connectTries := StrToInt(edtConnectTries.Text);
    if (connectTries <= 0) or (ckbAdvancedMode.Checked = false) then
       connectTries := defaultConnectTries;

    wireMode := bootloaderRoutines.connect(edtBootloaderPassword.Text, connectTries, pgbProgramConnect);
  except
    on E : Exception do
    begin
      logMessage('Verbindungsaufbau:' + sLineBreak + E.Message +
        sLineBreak + sLineBreak + 'Update-Vorgang wird abgebrochen', true, false, MB_ICONSTOP);
      Exit;
    end;
  end;

  if wireMode = 1 then
  begin
    oneWire := true;
    infoMessage('Verbunden: 1-Draht-Modus');
    logMessage('Achtung: Der 1-Draht-Modus wurde noch nicht getestet!', true, false, MB_ICONWARNING);
  end
  else
    infoMessage('Verbunden: 2-Draht-Modus');

  bootloaderRoutines.setWireMode(oneWire);

  //CRC-Unterstützung ermitteln
  supportsCRC := bootloaderRoutines.detectCRCSupport();
  bootloaderRoutines.resetCRC;

  if supportsCRC = true then
     infoMessage('Übertragung CRC-gesichert!')
  else
     infoMessage('CRC-Überprüfung nicht unterstützt');

  //Verify-Unterstützung ermitteln
  supportsVerify := bootloaderRoutines.detectVerifySupport();
  if supportsVerify = true then
     infoMessage('Überprüfung (Verify) unterstützt!')
  else
     infoMessage('Überprüfung (Verify) nicht unterstützt');

  Application.ProcessMessages;

  //Informationen auslesen
  try
    //Bootloader-Version
    deviceRevision := bootloaderRoutines.readRevision();
    if deviceRevision <> settingValidBootloader then
    begin
      raise Exception.Create('Ungültige Bootloader-Version erkannt: ' + deviceRevision +
        sLineBreak + sLineBreak + 'Ausschließlich Version ' + settingValidBootloader +
        ' wird unterstützt!');
      Exit;
    end;

    //Signatur
    deviceName := bootloaderRoutines.readSignature();

    if deviceName = defaultDeviceName then
    begin
      if showMessages = true then
      begin
        reply := Application.MessageBox('Es wurde ein unbekannter Mikrocontroller gefunden (unbekannte Signatur).' +
          sLineBreak + 'Das Firmware-Update könnte fehlschlagen und Daten verloren gehen.' +
          sLineBreak + sLineBreak + 'Update trotzdem durchführen?', 'UpdateLoader', MB_ICONWARNING or MB_YESNO or MB_DEFBUTTON2);

        if reply = IDNO then
          Exit;
      end
      else
        logMessage('Achtung: Unbekannter Mikrocontroller!');
    end;

    //Flash-Größe
    deviceFlashSize := bootloaderRoutines.readFlashSize();

    if deviceFlashSize > settingMaximumBufferSize then
    begin
      raise Exception.Create('Flash-Speicher größer als maximal mit diesem Programm ansprechbarer Bereich' +
        sLineBreak + sLineBreak + '(' +
        convertKB2String(deviceFlashSize) + ' Flash >>> ' +
        convertKB2String(settingMaximumBufferSize) + ' Maximum)');
      Exit;
    end;

    if deviceFlashSize < updateBufferUsed then
    begin
      raise Exception.Create('Flash-Speicher kleiner als Firmware-Datei' +
        sLineBreak + sLineBreak + '(' +
        IntToStr(deviceFlashSize) + ' Byte Flash <<< ' +
        IntToStr(updateBufferUsed) + ' Byte Firmware)');
      Exit;
    end;

    //Puffer-Größe
    deviceWriteBuffer := bootloaderRoutines.readBufferSize();

    if deviceWriteBuffer < settingMinimumWriteBuffer then
    begin
      raise Exception.Create('Sendepuffer zu klein' +
        sLineBreak + sLineBreak + '(' +
        IntToStr(deviceWriteBuffer) + ' Byte Puffer <<< ' +
        IntToStr(settingMinimumWriteBuffer) + ' Byte Minimum)');
      Exit;
    end;

    infoMessage('Bootloader V' + deviceRevision + ' auf ' + deviceName + ' gefunden');
    infoMessage('Flash-Größe: ' + convertKB2String(deviceFlashSize) + ' (Puffer: ' +
      IntToStr(deviceWriteBuffer) + ' Byte)');

  except
    on E : Exception do
    begin
      logMessage('Einstellungen auslesen:' + sLineBreak + E.Message +
        sLineBreak + sLineBreak + 'Update-Vorgang wird abgebrochen', true, false, MB_ICONSTOP);
      Exit;
    end;
  end;

  Application.ProcessMessages;

  //CRC zwischenprüfen
  if supportsCRC = true then
  begin
    try
      if bootloaderRoutines.checkCRC() = false then
      begin
        raise Exception.Create('Prüfsumme falsch, Übertragungsfehler!');
        Exit;
      end
      else
        infoMessage('CRC-Prüfsumme OK');
    except
      on E : Exception do
      begin
         logMessage('CRC-Prüfsumme:' + sLineBreak + E.Message +
           sLineBreak + sLineBreak + 'Update-Vorgang wird abgebrochen', true, false, MB_ICONSTOP);
         Exit;
      end;
    end;
  end;

  pgbUpdateProgress.StepIt;
  Application.ProcessMessages;

  flashNotice(true, 'Update läuft', 1000);

  //Programm aufspielen
  if programWrite = true then
  begin

    infoMessage('Programmiere ' + convertKB2String(updateBufferUsed) +
      ' von ' + convertKB2String(deviceFlashSize) + ' Flash...');

    try
      if bootloaderRoutines.writeFirmware(updateBuffer, updateBufferUsed, deviceWriteBuffer, pgbProgramWrite) = false then
      begin
        raise Exception.Create('Programmieren fehlgeschlagen');
        Exit;
      end
      else
         infoMessage('Programmieren erfolgreich');
    except
      on E: Exception do
      begin
         logMessage('Programmieren: (Adresse 0x' + bootloaderRoutines.getFailureAddress + ')' + sLineBreak + E.Message +
           sLineBreak + sLineBreak + 'Update-Vorgang wird abgebrochen' +
           sLineBreak + 'Achtung: Original-Firmware wurde möglicherweise überschrieben oder beschädigt!', true, false, MB_ICONSTOP);
         Exit;
      end;
    end;

    pgbUpdateProgress.StepIt;
  end
  else
    pgbUpdateProgress.Max := pgbUpdateProgress.Max - 10;

  Application.ProcessMessages;

  //Abfrage auf optionales Überprüfen
  if (showMessages = true) and (programVerify = false) and (supportsVerify = true) then
  begin
    reply := Application.MessageBox('Die geschriebenen Daten sollte zur Erhöhung der Funktionssicherheit überprüft werden.' +
      sLineBreak + 'In den erweiterten Einstellungen wurde die Überprüfung jedoch deaktiviert.' +
      sLineBreak + sLineBreak + 'Überprüfung jetzt trotzdem durchführen?', 'UpdateLoader', MB_ICONQUESTION or MB_YESNO or MB_DEFBUTTON1);

    if reply = IDYES then
      programVerify := true;
  end;

  //Programm prüfen
  if (programVerify = true) and (supportsVerify = true) then
  begin

    infoMessage('Überprüfe gespeicherte Firmware...');

    try
      if bootloaderRoutines.verifyFirmware(updateBuffer, updateBufferUsed, pgbProgramVerify) = false then
      begin
        raise Exception.Create('Überprüfung fehlgeschlagen');
        Exit;
      end
      else
         infoMessage('Überprüfung erfolgreich');
    except
      on E: Exception do
      begin
         logMessage('Programm überprüfen: (Adresse 0x' + bootloaderRoutines.getFailureAddress + ')' + sLineBreak + E.Message +
           sLineBreak + sLineBreak + 'Update-Vorgang wird abgebrochen', true, false, MB_ICONSTOP);
         Exit;
      end;
    end;

    pgbUpdateProgress.StepIt;
  end
  else
    pgbUpdateProgress.Max := pgbUpdateProgress.Max - 10;

  Application.ProcessMessages;

  //CRC prüfen
  if supportsCRC = true then
  begin
    try
      if bootloaderRoutines.checkCRC() = false then
      begin
        raise Exception.Create('Prüfsumme falsch, Übertragungsfehler!');
        Exit;
      end
      else
        infoMessage('CRC-Prüfsumme OK');
    except
      on E : Exception do
      begin
         logMessage('CRC-Prüfsumme:' + sLineBreak + E.Message +
           sLineBreak + sLineBreak + 'Update-Vorgang wird abgebrochen', true, false, MB_ICONSTOP);
         Exit;
      end;
    end;
  end;

  Application.ProcessMessages;

  //Programm starten
  if programStart = true then
  begin
    if bootloaderRoutines.startFirmware() = false then
    begin
       if showMessages = true then
       begin
          logMessage('Die neue Firmware konnte nicht automatisch gestartet werden. (Kommando wird nicht unterstützt)' +
            sLineBreak + sLineBreak + 'Gerät bitte manuell neu starten um die neue Firmware zu aktivieren!', true, false, MB_ICONINFORMATION);
       end
       else
          infoMessage('Firmware nicht gestartet (Kommando unbekannt)');
    end
    else
       infoMessage('Neue Firmware gestartet');

    pgbUpdateProgress.StepIt;
  end
  else
    pgbUpdateProgress.Max := pgbUpdateProgress.Max - 10;

  Application.ProcessMessages;

  infoMessage('Firmware-Update erfolgreich beendet!');
  pgbUpdateProgress.Position := pgbUpdateProgress.Max;
  sleepActive(1000);

  Result := true;
end;
{$ENDREGION}

end.

