(*#######################################################################################

UpdateLoader - Hilfsfunktionen

http://www.leo-andres.de/

Copyright (C) 2012 Leo-Andres Hofmann


This file is part of UpdateLoader.

UpdateLoader is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

UpdateLoader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with UpdateLoader. If not, see <http://www.gnu.org/licenses/>.

#######################################################################################*)

unit tools;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ComCtrls,
  Grids, ExtCtrls, EditBtn, Buttons, LCLType, LCLIntf, Registry;

type
  TMemoryBuffer = array of byte;
  TByteList = array of byte;

function getComportCount(): integer;
procedure getComportNames(ports: TStrings);
function getComportFriendlyName(regKey: string; port: string): string;

procedure showProgramMemory(memoryView: TStringGrid; buffer: TMemoryBuffer; bufferLength: longint; progressBar: TProgressBar = nil);
function loadProgramFile(fileName: string; buffer: TMemoryBuffer; bufferLength: longint; var usedBuffer: longint): boolean;
procedure clearBuffer(buffer: TMemoryBuffer);

function convertByteList2String(data: TByteList): string;
function convertByteList2HexString(data: TByteList): string;
function convertKB2String(kb: longint; appendUnit: boolean = true): string;
implementation

uses
  Protocol;

{$REGION 'Com-Ports auflisten'}
function getComportCount(): integer;
var
  registry: TRegistry;
  keyInfo: TRegKeyInfo;
begin
     registry := TRegistry.Create;
     Result := -1;

     try
       registry.RootKey := HKEY_LOCAL_MACHINE;
       if registry.OpenKeyReadOnly('HARDWARE\DEVICEMAP\SERIALCOMM') then
       begin
         registry.GetKeyInfo(keyInfo);
         Result := keyInfo.NumValues;
       end;
     finally
       registry.CloseKey;
       registry.Free;
     end;
end;

procedure getComportNames(ports: TStrings);
var
  registry: TRegistry;
  portNames: TStringList;
  portList: TStringList;
  count: integer;
  port: string;
  friendlyName: string;
begin
     registry := TRegistry.Create;
     portNames := TStringList.Create;
     portList := TStringList.Create;

     portNames.Clear;
     portList.Clear;

     ports.BeginUpdate;
     try
       ports.Clear;

       registry.RootKey := HKEY_LOCAL_MACHINE;
       if registry.OpenKeyReadOnly('HARDWARE\DEVICEMAP\SERIALCOMM') then
       begin
         registry.GetValueNames(portList);

         for count := 0 to portList.Count - 1 do
         begin
           port := registry.ReadString(portList[count]);
           friendlyName := getComportFriendlyName('\SYSTEM\CurrentControlSet\Enum\', port);

           if friendlyName <> '' then
             portNames.Add(port + ': ' + friendlyName);
         end;

         portNames.Sort;
         ports.Assign(portNames);
       end;
     finally
       registry.CloseKey;
       registry.Free;
       portNames.Free;
       portList.Free;

       ports.EndUpdate;
     end;
end;

function getComportFriendlyName(regKey: string; port: string): string;
var
  registry: TRegistry;
  keyNames: TStringList;
  count: integer;
  currentKey: string;
  friendlyName: string;
  foundKeyIn: string;
begin
     registry := TRegistry.Create;
     keyNames := TStringList.Create;
     Result := '';

     registry.RootKey := HKEY_LOCAL_MACHINE;
     registry.OpenKeyReadOnly(regKey);
     registry.GetKeyNames(keyNames);
     registry.CloseKey;

     try
       for count := 0 to keyNames.Count - 1 do
       begin
         currentKey := regKey + keyNames[count] + '\';

         if registry.OpenKeyReadOnly(currentKey + 'Device Parameters') then
         begin
           if registry.ReadString('PortName') = port then
           begin
             registry.CloseKey;
             registry.OpenKeyReadOnly(currentKey);
             friendlyName := registry.ReadString('FriendlyName');
             registry.CloseKey;

             foundKeyIn := copy(currentKey, 32, length(currentKey) - 32);
             foundKeyIn := copy(foundKeyIn, 0, pos('\', foundKeyIn) - 1);

             friendlyName := friendlyName + ' - (' + foundKeyIn + ')';

             if (settingHideACPIPorts = true) and (foundKeyIn = 'ACPI') then
               friendlyName := '';

             Break;
           end;
         end
         else
         begin
           if registry.OpenKeyReadOnly(currentKey) and registry.HasSubKeys then
           begin
             friendlyName := getComportFriendlyName(currentKey, port);
             registry.CloseKey;
             if friendlyName <> '' then Break;
           end;
         end;
       end;

       Result := friendlyName;
     finally
       registry.CloseKey;
       registry.Free;
       keyNames.Free;
     end;
end;
{$ENDREGION}

{$REGION 'Hex-Datei anzeigen und auslesen'}
procedure showProgramMemory(memoryView: TStringGrid; buffer: TMemoryBuffer; bufferLength: longint; progressBar: TProgressBar = nil);
var
  count: longint;
  lineCount, byteCount: integer;
  dataHex, dataChr, adressChr: string;
begin
  lineCount := 0;
  byteCount := 1;
  dataHex := '';
  dataChr := '';

  if Assigned(progressBar) then
  begin
    progressBar.Position := 0;
    progressBar.Max := Round(bufferLength / 800);
    progressBar.Step := 1;
  end;

  memoryView.Hide;
  memoryView.RowCount := 2;

  memoryView.ColWidths[0] := 55;
  memoryView.ColWidths[1] := 170;
  memoryView.ColWidths[2] := 15;
  memoryView.ColWidths[3] := 175;
  memoryView.ColWidths[4] := 120;

  memoryView.Cells[0, 0] := 'Adresse';
  memoryView.Cells[1, 0] := 'Low-Bytes';
  memoryView.Cells[3, 0] := 'High-Bytes';
  memoryView.Cells[4, 0] := 'Daten';

  for count := low(buffer) to bufferLength do
  begin

    dataHex := dataHex + IntToHex(buffer[count], 2) + ' ';

    if buffer[count] <> 0 then
       dataChr := dataChr + AnsiChar(buffer[count])
    else
        dataChr := dataChr + ' ';

    if (byteCount = 16) or (count = bufferLength) then
    begin
      adressChr := IntToHex(count - 15, 6);

      memoryView.Cells[0, memoryView.RowCount - 1] := adressChr;
      memoryView.Cells[4, memoryView.RowCount - 1] := dataChr;

      if Length(dataHex) >= 24 then
      begin
        memoryView.Cells[1, memoryView.RowCount - 1] := copy(dataHex, 1, 24);
        memoryView.Cells[3, memoryView.RowCount - 1] := copy(dataHex, 25, Length(dataHex) - 1);
        memoryView.Cells[2, memoryView.RowCount - 1] := '-';
      end
      else
      begin
        memoryView.Cells[1, memoryView.RowCount - 1] := copy(dataHex, 1, Length(dataHex) - 1);
      end;

      memoryView.RowHeights[memoryView.RowCount -1] := 18;
      memoryView.RowCount := memoryView.RowCount + 1;

      dataHex := '';
      dataChr := '';
      byteCount := 0;

      inc(lineCount);
      if lineCount >= 50 then
      begin
        if Assigned(progressBar) then
           progressBar.StepIt;

        Application.ProcessMessages;
        lineCount := 0;
      end;
    end;

    inc(byteCount);
  end;

  memoryView.RowCount := memoryView.RowCount - 1;
  memoryView.Show;

  if Assigned(progressBar) then
     progressBar.Position := 0;
end;

{$R-}
function loadProgramFile(fileName: string; buffer: TMemoryBuffer; bufferLength: longint; var usedBuffer: longint): boolean;
var
  programFile: textfile;
  readLine, temp: string;
  data, checksum, byteCount, dataLength: byte;
  address: word;
  memoryAddress, segmentAddress: longint;
  segmentedAdress: boolean;
  recordType, lineChecksum: byte;
begin
  checksum := 0;
  lineChecksum := 0;
  segmentAddress := 0;
  segmentedAdress := false;
  Result := false;

  clearBuffer(buffer);
  usedBuffer := 0;

  try
    AssignFile(programFile, fileName);
    Reset(programFile);

    while not EOF(programFile) do
    begin
      ReadLn(programFile, readLine);

      if readLine[1] = ':' then
      begin
        checksum := 0;

        //Datensatz-Länge
        temp := '$' + copy(readLine, 2, 2);
        dataLength := StrToInt(temp);
        checksum := checksum + dataLength;

        //Adresse
        temp := '$' + copy(readLine, 4, 4);
        address := StrToInt(temp);
        checksum := checksum + lo(address) + hi(address);

        if segmentedAdress = true then
           memoryAddress := address + segmentAddress
        else
           memoryAddress := address;

        //Record-Typ
        temp := '$' + copy(readLine, 8, 2);
        recordType := StrToInt(temp);
        checksum := checksum + recordType;

        //Record-Typen verarbeiten
        case recordType of
          0: //Datenblock
            begin
              if (memoryAddress + dataLength - 1) <= bufferLength then
              begin
                byteCount := 0;

                while byteCount < dataLength do
                begin
                  temp := '$' + copy(readLine, 10 + 2 * byteCount, 2);
                  data := StrToInt(temp);
                  checksum := checksum + data;

                  //Doppelte Einträge abfangen
                  if buffer[memoryAddress + byteCount] <> StrToInt('$FF') then
                  begin
                    if buffer[memoryAddress + byteCount] = data then
                    begin
                      raise Exception.Create('Doppelter Eintrag in Datei, Speicherstelle überschrieben');
                      break;
                    end;
                  end;

                  buffer[memoryAddress + byteCount] := data;
                  inc(byteCount);
                end;

                //temp := '$' + copy(readLine, 10 + 2 * byteCount, 2);
                temp := '$' + RightStr(readLine, 2);
                lineChecksum := StrToInt(temp);

                //Belegten Speicher zählen
                if usedBuffer <= memoryAddress + byteCount - 1 then
                   usedBuffer := memoryAddress + byteCount - 1;
              end
              else
              begin
                raise Exception.Create('Firmware größer als Pufferspeicher (' +
                      convertKB2String(bufferLength) + ')');
                break;
              end;
            end;
          1: //Dateiende
            begin
              temp := '$' + copy(readLine, 10, 2);
              lineChecksum := StrToInt(temp);
            end;
          2: //Segment-Adresse
            begin
              segmentedAdress := true;

              temp := '$' + copy(readLine, 10, 4);
              segmentAddress := StrToInt(temp) * 16;

              temp := '$' + copy(readLine, 14, 2);
              lineChecksum := StrToInt(temp);
            end;
          else
            begin
              raise Exception.Create('Ungültiger Record (unbekannte Funktion)');
              break;
            end;
        end;

        //Prüfsumme
        checksum := lineChecksum + checksum;
        if ((checksum and $00FF) <> 0) then
        begin
          raise Exception.Create('Prüfsumme der Firmware-Datei ungültig');
          break;
        end;
      end
      else
      begin
        raise Exception.Create('Ungültiges Dateiformat (kein Intel-Hex)');
        break;
      end;

      Application.ProcessMessages;

    end;

    if usedBuffer <= 0 then
       raise Exception.Create('Firmware-Datei enthält keine Einträge');
  finally
    CloseFile(programFile);
  end;

  Result := true;
end;
{$R+}

procedure clearBuffer(buffer: TMemoryBuffer);
var
  count: longint;
begin
  for count := low(buffer) to high(buffer) do
      buffer[count] := $FF;
end;
{$ENDREGION}

{$REGION 'Formatumwandlungen'}

function convertByteList2String(data: TByteList): string;
var
  count: integer;
  convert: string;
begin
  convert := '';
  Result := '';

  for count := low(data) to high(data) do
      convert := convert + IntToStr(data[count]);

  Result := convert;
end;

function convertByteList2HexString(data: TByteList): string;
var
  count: integer;
  convert: string;
begin
  convert := '';
  Result := '';

  for count := low(data) to high(data) do
      convert := convert + IntToHex(data[count], 2);

  Result := convert;
end;

function convertKB2String(kb: longint; appendUnit: boolean = true): string;
var
  convert: string;
begin
  convert := '';
  Result := '';

  if kb >= 1024 then
  begin
    convert := IntToStr(Round(kb / 1024));
    if appendUnit = true then
      convert := convert + 'kB';
  end
  else
  begin
    convert := IntToStr(kb);
    if appendUnit = true then
      convert := convert + ' Byte';
  end;

  Result := convert;
end;
{$ENDREGION}

end.

